require 'zlib'
require 'wiki_xml_parser'

module Openable
  def self.create(*args)
    puts "Open\n"
    obj = self.new(args)
    begin
      yield obj
    ensure
      obj.close if obj.respond_to?(:close)
    end
  end

  public_class_method :create
end

class DumpWriter
  Default_Max_Size = 1536 * 1024 * 1024
  
  def initialize(prefix, options = {:compress => false, :max_file_size => DumpWriter::Default_Max_Size})
    puts "pre: #{prefix}, max: #{options[:max_file_size]}"
    options[:max_file_size] = DumpWriter::Default_Max_Size unless options[:max_file_size]
    
    @prefix, @options = prefix, options
    
    @compress, @max_file_size = options[:compress], options[:max_file_size]
    @current_size = -1
    @files = 0
  end
  
  def self.open(*args)
    obj = self.new(*args)
    begin
      yield obj
    ensure
      obj.close if obj.respond_to?(:close)
    end
  end
  
  def close
    close_current_stream if @file
    @current_size = -2
  end
  
  def close_current_stream
    @current_size = -1
    @file.close
    @file = nil
  end
  
  def open_new_stream
    raise "Unexpected open_new_stream()" unless @current_size == -1
    
    @current_size = 0
    @file = File.new("#{@prefix}.#{@files}", 'wb')
    @files += 1
    Marshal.dump(@options, @file) if @files == 1
  end
  
  def offset; [@files - 1, @current_size] end
  def offset_str; o = offset; "#{o[0]},#{o[1]}" end
  
  def write(page)
    data = Marshal.dump(page)
    
    if @compress
      z = Zlib::Deflate.new(Zlib::BEST_COMPRESSION, -Zlib::MAX_WBITS)
      data = z.deflate(data, Zlib::FINISH)
    end

    #    data = Zlib::Deflate.deflate(data) if @compress
    #    puts "#{data.length}"
    
    if @current_size + data.length > @max_file_size
      close_current_stream
    end
    
    open_new_stream if @current_size < 0
    
    @file.write(data)
    @current_size += data.length
  end
  alias << write
end

class DumpReader
  def initialize(prefix)
    @buffer = nil
    @current_file = -1
    @prefix = prefix
  end

  def close
    @file = nil
  end

  def self.open(*args)
    obj = self.new(*args)
    begin
      yield obj
    ensure
      obj.close if obj.respond_to?(:close)
    end
  end
  
  def open_next_stream
    @file.close if @file
    
    @current_file += 1
    
    filename = "#{@prefix}.#{@current_file}"
    return false unless File::exist?(filename)
    @file = File.new(filename, 'rb')
    
    @options = Marshal.load(@file) if @current_file == 0
    true
  end
  
  def check_stream
    if @buffer.nil? and (@file.nil? or @file.eof?)
      open_next_stream
    else
      true
    end
  end
  
  def read
    return nil unless check_stream
    #    open_next_stream if @file.nil? or @file.eof?
    
    if @options[:compress]
      inflater = Zlib::Inflate.new(Zlib::MAX_WBITS)
      
      out_buffer = []
      while true
        @buffer = @file.read(64 * 1024) if @buffer.nil?
        
        #        puts "#{@buffer.length}\n"
        
        old_total_in = inflater.total_in

        begin
          out_buffer << inflater.inflate(@buffer)
        rescue Zlib::BufError => e
          #          @buffer = @buffer + @file.read(1024)
          puts "#{inflater.total_in}  #{inflater.total_out}   #{@buffer.length}\n"
          #          retry
          #          raise e
          error = true
        end
        
        bytes_read_from_buffer = inflater.total_in - old_total_in
        if bytes_read_from_buffer < @buffer.length
          @buffer = @buffer[bytes_read_from_buffer..@buffer.length]
        else
          @buffer = nil
        end
        
        break if inflater.finished?
      end
      
      data = out_buffer.join
    else
      data = @file
    end
    #    puts "#{data.length}"

    if error
      puts "#{data}\n"
      d = Marshal.load(data)
      puts "#{d[:text]}"
      pp d
      #      raise "sdfasf"
    end
    
    Marshal.load(data)
  end
  alias >> read
end

class ReadBinaryDump < DumpReader
  def initialize
    super "#{Config[:root]}dump/chunk"
  end
  
  def run
    item = read
    while item
      yield item
      item = read
    end
  end
  
  def finish; close end
end

class WriteBinaryDump < DumpWriter
  def initialize
    super "#{Config[:root]}dump/chunk"
  end
  
  def finish; close end
end

class ReadXml < WikiXmlParser
  def initialize
    super Config[:xml_dump]
  end
  
  def finish
    File.open("#{Config[:root]}namespaces.dat", 'wb') do |f|
      Marshal.dump(namespaces_by_key, f)
      Marshal.dump(namespaces_by_name, f)
    end
  end
end

if __FILE__ == $0
  DumpReader.open('e:/wiki/dump/chunk') do |dr|
    while true
      hash = dr.read
      #      pp hash
      break if hash.nil?
    end
  end
  
  if 1 == 0
    DumpWriter.open('c:/temp/test/dump_file', :max_file_size => 1024, :compress => false) do |pw|
      (1..1024).each do |i|
        hash = {:num => i}
        pw << hash
        puts "#{pw.offset}:::#{pw.offset_str}\n"
      end
    end
  
    DumpReader.open('c:/temp/test/dump_file') do |dr|
      (1..1025).each do |i|
        hash = dr.read
        p hash
      end
    end
  end
end
