require 'WikiMiner/db'

class Logger
  def initialize(*args)
    @targets = []
    args.each do |arg|
      arg.each do |key, value|
        case key
        when :file
          @targets << File.new(value, 'a')
        when :stdout
          @targets << $stdout
        end
      end
    end
  end
  
  def flush
    @targets.each { |target| target.flush }
  end
  
  def log(message = '')
    log_message = "#{Time.now}: #{message}\n"
    
    @targets.each do |target|
      target.puts log_message
    end
  end
  alias << log
  
  def newline
    log
  end
end

class TaskManager
  def initialize(options)
    @options = options
    @objs = {}
    @stages = []

    @log = Logger.new(:file => Config[:log], :stdout => true)
  end

  def bind(binds = {})
    binds.each { |k, v| @objs[k] = v }
  end

  def stage(source, *tasks)
    @stages << {:source => source, :tasks => tasks}
  end
  def task(task, &block); [task, block] end

  def lock
    begin
      File.unlink(Config[:lock]) if File.exists?(Config[:lock])
    rescue
    end
    
    raise Exception, "TaskManager is locked." if File.exists?(Config[:lock])
    
    File.open(Config[:lock], 'w') { yield }
    File.unlink(Config[:lock])
  end
  
  def is_in_list(param, value)
    if @options[param]
      if @options[param].respond_to?(:include?)
        @options[param].include?(value)
      else
        @options[param] == value
      end
    end
  end
  
  def active_task?(name)
    v = is_in_list(:only, name)
    
    if v
      task = Task.find(:first, :conditions => {:name => name.to_s})
      v = task.nil? || task.status == 0
    end

    return v
    #    return v
  end
  
  class ExecutionTime
    def initialize
      @start_time = Time.now
    end
    
    def to_s
      return (Time.now - @start_time).to_s
    end
  end
  
  def time(&block)
    block.call(ExecutionTime.new)
    #    start_time = Time.now
    #    t = lambda { Time.now - start_time }
    #    block.call(t)
  end
  
  def get_object(name)
    obj = @objs[name]
    if obj.nil?
      name = name.to_s unless name.is_a?(String)
      name = name.to_s.split('_').map { |s| s.capitalize }.join
      c = Module.const_get(name)
      obj = c.new if c.is_a?(Class)
    end
    
    return obj
  end

  def run
    lock do
      begin
        time_all = ExecutionTime.new
        @log << "Task manager created."
        @log << "#{@stages.length} stages and #{@objs.length} objects registered."
        @log.newline
        
        active_stages = []
        @stages.each do |stage|
          active_tasks = []

          stage[:tasks].each do |task, block|
            active_tasks << {:name => task, :handler => get_object(task)} if active_task?(task)
          end
          
          active_stages << {:source => stage[:source], :tasks => active_tasks} if !active_tasks.empty?
        end

        
        active_stages.each do |stage|
          Config[:db].transaction do
            time_task = ExecutionTime.new
            @log << "Executing stage from source '#{stage[:source]}' with #{stage[:tasks].length} active tasks."
            
            handlers = stage[:tasks].collect { |t| t[:handler] }
            
            source = get_object(stage[:source])
            source.run do |data|
              handlers.each do |task|
                task << data
              end
            end

            handlers.each do |task, _|
              task.finish if task.respond_to?(:finish)
            end
            source.finish if source.respond_to?(:finish)
              
            @log << "Finished tasks in stage #{stage[:source]} in #{time_task} secs."
            @log.newline
          end
         
          stage[:tasks].each do |t|
            task_name = t[:name].to_s
            task = Task.find(:first, :conditions => {:name => task_name}) || Task.new(:name => task_name)
            task.status = 1
            task.save
          end
        end

        @log << "All tasks finished in #{time_all} secs."
      ensure
        @log.flush
      end
    end
  end

  def self.default(*args)
    @options ||= {}
    @options.clear
    args.each { |arr| arr.each { |key, value| @options[key] = value } }
  end

  def self.register(&block)
    @instance ||= TaskManager.new(@options)
    @instance.instance_eval(&block)
  end
  
  def self.run
    @instance.run if @instance
  end
  
  private :initialize
end
